"""Tests for project CRUD endpoints."""


class TestCreateProject:
    def test_create_project(self, client, tmp_path):
        project_dir = tmp_path / "my-project"
        project_dir.mkdir()
        resp = client.post("/api/projects", json={
            "display_name": "My Project",
            "working_dir": str(project_dir),
        })
        assert resp.status_code == 201
        data = resp.json()
        assert data["display_name"] == "My Project"
        assert data["dir_exists"] is True
        assert "id" in data
        assert "created_at" in data

    def test_resolves_path(self, client, tmp_path):
        sub = tmp_path / "sub"
        sub.mkdir()
        # Pass a path with ".." — should be resolved
        unresolved = str(tmp_path / "sub" / ".." / "sub")
        resp = client.post("/api/projects", json={
            "display_name": "Test",
            "working_dir": unresolved,
        })
        assert resp.status_code == 201
        assert ".." not in resp.json()["working_dir"]

    def test_duplicate_working_dir_returns_409(self, client, tmp_path):
        project_dir = tmp_path / "dup"
        project_dir.mkdir()
        client.post("/api/projects", json={
            "display_name": "First",
            "working_dir": str(project_dir),
        })
        resp = client.post("/api/projects", json={
            "display_name": "Second",
            "working_dir": str(project_dir),
        })
        assert resp.status_code == 409


class TestListProjects:
    def test_empty_list(self, client):
        resp = client.get("/api/projects")
        assert resp.status_code == 200
        assert resp.json() == []

    def test_lists_created_projects(self, client, tmp_path):
        d1 = tmp_path / "p1"
        d1.mkdir()
        d2 = tmp_path / "p2"
        d2.mkdir()
        client.post("/api/projects", json={"display_name": "P1", "working_dir": str(d1)})
        client.post("/api/projects", json={"display_name": "P2", "working_dir": str(d2)})

        resp = client.get("/api/projects")
        projects = resp.json()
        assert len(projects) == 2
        assert projects[0]["display_name"] == "P1"
        assert projects[1]["display_name"] == "P2"

    def test_stale_detection(self, client, tmp_path):
        resp = client.post("/api/projects", json={
            "display_name": "Ghost",
            "working_dir": str(tmp_path / "nonexistent"),
        })
        assert resp.status_code == 201

        resp = client.get("/api/projects")
        projects = resp.json()
        assert len(projects) == 1
        assert projects[0]["dir_exists"] is False


class TestGetProject:
    def test_get_existing(self, client, tmp_path):
        d = tmp_path / "proj"
        d.mkdir()
        create_resp = client.post("/api/projects", json={
            "display_name": "Test",
            "working_dir": str(d),
        })
        pid = create_resp.json()["id"]

        resp = client.get(f"/api/projects/{pid}")
        assert resp.status_code == 200
        assert resp.json()["display_name"] == "Test"
        assert resp.json()["dir_exists"] is True

    def test_get_missing_returns_404(self, client):
        resp = client.get("/api/projects/999")
        assert resp.status_code == 404


class TestUpdateProject:
    def test_update_display_name(self, client, tmp_path):
        d = tmp_path / "proj"
        d.mkdir()
        create_resp = client.post("/api/projects", json={
            "display_name": "Old Name",
            "working_dir": str(d),
        })
        pid = create_resp.json()["id"]

        resp = client.patch(f"/api/projects/{pid}", json={
            "display_name": "New Name",
        })
        assert resp.status_code == 200
        assert resp.json()["display_name"] == "New Name"

    def test_update_missing_returns_404(self, client):
        resp = client.patch("/api/projects/999", json={"display_name": "X"})
        assert resp.status_code == 404


class TestDeleteProject:
    def test_delete_project(self, client, tmp_path):
        d = tmp_path / "proj"
        d.mkdir()
        create_resp = client.post("/api/projects", json={
            "display_name": "Doomed",
            "working_dir": str(d),
        })
        pid = create_resp.json()["id"]

        resp = client.delete(f"/api/projects/{pid}")
        assert resp.status_code == 200

        # Verify gone
        resp = client.get(f"/api/projects/{pid}")
        assert resp.status_code == 404

    def test_delete_missing_returns_404(self, client):
        resp = client.delete("/api/projects/999")
        assert resp.status_code == 404

    def test_delete_imu_project_returns_403(self, client):
        """Bot Imu virtual project (id=0) cannot be deleted."""
        resp = client.delete("/api/projects/0")
        assert resp.status_code == 403
        assert "Bot Imu" in resp.json()["detail"]



"""Unit tests for SSE utilities and TreeState builder."""

from strawpot_gui.sse import TreeState, format_sse, sse_retry


class TestFormatSSE:
    def test_basic_format(self):
        result = format_sse(1, {"hello": "world"})
        assert result == 'id: 1\ndata: {"hello":"world"}\n\n'

    def test_retry_format(self):
        result = sse_retry(5000)
        assert result == "retry: 5000\n\n"

    def test_default_retry(self):
        result = sse_retry()
        assert result == "retry: 3000\n\n"


class TestTreeStateSessionJson:
    def test_load_root_agent(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "strawpot-claude-code",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        result = state.to_dict()
        assert len(result["nodes"]) == 1
        assert result["nodes"][0]["agent_id"] == "agent_root"
        assert result["nodes"][0]["status"] == "running"

    def test_load_ignores_existing_agents(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "strawpot-claude-code",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        # Loading again shouldn't overwrite
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "different",
                    "runtime": "different",
                    "parent": None,
                    "started_at": "T1",
                }
            }
        })
        assert state.to_dict()["nodes"][0]["role"] == "orchestrator"

    def test_empty_agents(self):
        state = TreeState()
        state.load_session_json({"agents": {}})
        assert state.to_dict()["nodes"] == []


class TestTreeStateTraceEvents:
    def test_delegate_start_creates_pending(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "cc",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        state.process_event({
            "event": "session_start", "span_id": "s0", "data": {},
        })
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": "s0",
            "data": {"role": "implementer"},
        })
        result = state.to_dict()
        assert len(result["pending_delegations"]) == 1
        assert result["pending_delegations"][0]["role"] == "implementer"
        assert result["pending_delegations"][0]["requested_by"] == "agent_root"

    def test_agent_spawn_promotes_pending(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "cc",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        state.process_event({
            "event": "session_start", "span_id": "s0", "data": {},
        })
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": "s0",
            "data": {"role": "implementer"},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s1", "ts": "T2",
            "data": {"agent_id": "agent_impl", "runtime": "cc", "pid": 999},
        })
        result = state.to_dict()
        assert len(result["pending_delegations"]) == 0
        agent_ids = [n["agent_id"] for n in result["nodes"]]
        assert "agent_impl" in agent_ids
        impl = next(n for n in result["nodes"] if n["agent_id"] == "agent_impl")
        assert impl["parent"] == "agent_root"
        assert impl["status"] == "running"

    def test_agent_end_marks_completed(self):
        state = TreeState()
        state.process_event({
            "event": "session_start", "span_id": "s0", "data": {},
        })
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": "s0",
            "data": {"role": "impl"},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s1", "ts": "T1",
            "data": {"agent_id": "a1", "runtime": "cc", "pid": 1},
        })
        state.process_event({
            "event": "agent_end", "span_id": "s1",
            "data": {"exit_code": 0, "duration_ms": 5000},
        })
        node = state.nodes["a1"]
        assert node.status == "completed"
        assert node.exit_code == 0
        assert node.duration_ms == 5000

    def test_agent_end_nonzero_marks_failed(self):
        state = TreeState()
        state.process_event({
            "event": "delegate_start", "span_id": "s1",
            "parent_span": None,
            "data": {"role": "impl"},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s1", "ts": "T1",
            "data": {"agent_id": "a1", "runtime": "cc", "pid": 1},
        })
        state.process_event({
            "event": "agent_end", "span_id": "s1",
            "data": {"exit_code": 1, "duration_ms": 3000},
        })
        assert state.nodes["a1"].status == "failed"

    def test_delegate_denied_recorded(self):
        state = TreeState()
        state.process_event({
            "event": "delegate_denied", "span_id": "s2",
            "parent_span": "s0",
            "data": {"role": "admin", "reason": "DENY_DEPTH_LIMIT"},
        })
        result = state.to_dict()
        assert len(result["denied_delegations"]) == 1
        assert result["denied_delegations"][0]["role"] == "admin"
        assert result["denied_delegations"][0]["reason"] == "DENY_DEPTH_LIMIT"

    def test_agent_spawn_uses_event_role_when_no_pending(self):
        """Root agent_spawn (no prior delegate_start) reads role from event data."""
        state = TreeState()
        state.process_event({
            "event": "session_start", "span_id": "s0", "data": {},
        })
        state.process_event({
            "event": "agent_spawn", "span_id": "s0", "ts": "T0",
            "data": {
                "agent_id": "agent_root",
                "role": "ai-ceo",
                "runtime": "strawpot-claude-code",
                "pid": 123,
            },
        })
        assert state.nodes["agent_root"].role == "ai-ceo"

    def test_session_end_marks_terminal(self):
        state = TreeState()
        state.process_event({
            "event": "session_end", "span_id": "s0",
            "data": {"duration_ms": 1000},
        })
        assert state.is_terminal

    def test_session_end_completes_root(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "cc",
                    "parent": None,
                    "started_at": "T0",
                }
            }
        })
        state.process_event({
            "event": "session_end", "span_id": "s0",
            "data": {"duration_ms": 5000},
        })
        assert state.nodes["agent_root"].status == "completed"
        assert state.nodes["agent_root"].duration_ms == 5000


class TestTreeStateFullLifecycle:
    def test_full_scenario(self):
        state = TreeState()
        state.load_session_json({
            "agents": {
                "agent_root": {
                    "role": "orchestrator",
                    "runtime": "strawpot-claude-code",
                    "parent": None,
                    "started_at": "T0",
                    "pid": 100,
                }
            }
        })
        events = [
            {"event": "session_start", "span_id": "s0",
             "data": {"run_id": "r1", "role": "orchestrator"}},
            {"event": "delegate_start", "span_id": "s1", "parent_span": "s0",
             "data": {"role": "implementer"}},
            {"event": "agent_spawn", "span_id": "s1", "ts": "T1",
             "data": {"agent_id": "agent_impl", "runtime": "cc", "pid": 200}},
            {"event": "delegate_start", "span_id": "s2", "parent_span": "s1",
             "data": {"role": "reviewer"}},
            {"event": "delegate_denied", "span_id": "s3", "parent_span": "s1",
             "data": {"role": "admin", "reason": "DENY_DEPTH_LIMIT"}},
            {"event": "agent_spawn", "span_id": "s2", "ts": "T2",
             "data": {"agent_id": "agent_rev", "runtime": "cc", "pid": 300}},
            {"event": "agent_end", "span_id": "s2",
             "data": {"exit_code": 0, "duration_ms": 10000}},
            {"event": "delegate_end", "span_id": "s2",
             "data": {"exit_code": 0,
                       "duration_ms": 10000}},
            {"event": "agent_end", "span_id": "s1",
             "data": {"exit_code": 0, "duration_ms": 30000}},
            {"event": "delegate_end", "span_id": "s1", "parent_span": None,
             "data": {"exit_code": 0,
                       "duration_ms": 30000}},
            {"event": "session_end", "span_id": "s0",
             "data": {"duration_ms": 35000}},
        ]
        for ev in events:
            state.process_event(ev)

        result = state.to_dict()

        # 3 nodes: root, implementer, reviewer
        assert len(result["nodes"]) == 3
        ids = {n["agent_id"] for n in result["nodes"]}
        assert ids == {"agent_root", "agent_impl", "agent_rev"}

        # All completed
        for node in result["nodes"]:
            assert node["status"] == "completed"

        # Reviewer's parent is implementer
        rev = next(n for n in result["nodes"] if n["agent_id"] == "agent_rev")
        assert rev["parent"] == "agent_impl"

        # Implementer's parent is root
        impl = next(
            n for n in result["nodes"] if n["agent_id"] == "agent_impl"
        )
        assert impl["parent"] == "agent_root"

        # No pending delegations
        assert len(result["pending_delegations"]) == 0

        # 1 denied delegation
        assert len(result["denied_delegations"]) == 1
        assert result["denied_delegations"][0]["role"] == "admin"

        assert state.is_terminal



"""Scheduled tasks CRUD endpoints."""

import sqlite3
from datetime import datetime, timezone

from croniter import croniter
from fastapi import APIRouter, Depends, HTTPException, Query
from pydantic import BaseModel, field_validator

from strawpot_gui.db import get_db_conn

router = APIRouter(prefix="/api", tags=["schedules"])


# ---------------------------------------------------------------------------
# Models
# ---------------------------------------------------------------------------


class ScheduleCreate(BaseModel):
    name: str
    project_id: int
    task: str
    cron_expr: str
    role: str | None = None
    system_prompt: str | None = None
    skip_if_running: bool = True
    conversation_id: int | None = None

    @field_validator("name")
    @classmethod
    def name_nonempty(cls, v: str) -> str:
        if not v.strip():
            raise ValueError("name must be non-empty")
        return v.strip()

    @field_validator("task")
    @classmethod
    def task_nonempty(cls, v: str) -> str:
        if not v.strip():
            raise ValueError("task must be non-empty")
        return v.strip()

    @field_validator("cron_expr")
    @classmethod
    def cron_valid(cls, v: str) -> str:
        v = v.strip()
        if not croniter.is_valid(v):
            raise ValueError(f"Invalid cron expression: {v}")
        return v


class OneTimeScheduleCreate(BaseModel):
    name: str
    project_id: int
    task: str
    run_at: str
    role: str | None = None
    system_prompt: str | None = None
    conversation_id: int | None = None

    @field_validator("name")
    @classmethod
    def name_nonempty(cls, v: str) -> str:
        if not v.strip():
            raise ValueError("name must be non-empty")
        return v.strip()

    @field_validator("task")
    @classmethod
    def task_nonempty(cls, v: str) -> str:
        if not v.strip():
            raise ValueError("task must be non-empty")
        return v.strip()

    @field_validator("run_at")
    @classmethod
    def run_at_valid(cls, v: str) -> str:
        v = v.strip()
        try:
            dt = datetime.fromisoformat(v)
        except ValueError:
            raise ValueError(f"Invalid ISO datetime: {v}")
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        if dt <= datetime.now(timezone.utc):
            raise ValueError("run_at must be in the future")
        return dt.isoformat()


class ScheduleUpdate(BaseModel):
    name: str | None = None
    task: str | None = None
    cron_expr: str | None = None
    run_at: str | None = None
    role: str | None = None
    system_prompt: str | None = None
    skip_if_running: bool | None = None
    conversation_id: int | None = None

    @field_validator("cron_expr")
    @classmethod
    def cron_valid(cls, v: str | None) -> str | None:
        if v is not None:
            v = v.strip()
            if not croniter.is_valid(v):
                raise ValueError(f"Invalid cron expression: {v}")
        return v

    @field_validator("run_at")
    @classmethod
    def run_at_valid(cls, v: str | None) -> str | None:
        if v is not None:
            v = v.strip()
            try:
                dt = datetime.fromisoformat(v)
            except ValueError:
                raise ValueError(f"Invalid ISO datetime: {v}")
            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=timezone.utc)
            if dt <= datetime.now(timezone.utc):
                raise ValueError("run_at must be in the future")
            return dt.isoformat()
        return v


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _row_to_dict(row: sqlite3.Row) -> dict:
    d = dict(row)
    d["enabled"] = bool(d.get("enabled", 0))
    d["skip_if_running"] = bool(d.get("skip_if_running", 1))
    return d


def _compute_next_run(cron_expr: str) -> str | None:
    try:
        now = datetime.now(timezone.utc)
        return croniter(cron_expr, now).get_next(datetime).isoformat()
    except (ValueError, KeyError):
        return None


# ---------------------------------------------------------------------------
# Endpoints
# ---------------------------------------------------------------------------


@router.get("/schedules")
def list_schedules(
    type: str | None = None, conn=Depends(get_db_conn),
):
    """List scheduled tasks with project names, optionally filtered by type."""
    if type and type not in ("recurring", "one_time"):
        raise HTTPException(422, "type must be 'recurring' or 'one_time'")

    query = (
        "SELECT s.*, p.display_name AS project_name "
        "FROM scheduled_tasks s "
        "LEFT JOIN projects p ON s.project_id = p.id "
    )
    params: list = []
    if type:
        query += "WHERE s.schedule_type = ? "
        params.append(type)
    query += "ORDER BY s.created_at DESC"

    rows = conn.execute(query, params).fetchall()
    return [_row_to_dict(row) for row in rows]


@router.post("/schedules", status_code=201)
def create_schedule(body: ScheduleCreate, conn=Depends(get_db_conn)):
    """Create a new scheduled task."""
    # Verify project exists
    project = conn.execute(
        "SELECT id FROM projects WHERE id = ?", (body.project_id,)
    ).fetchone()
    if not project:
        raise HTTPException(404, "Project not found")

    next_run = _compute_next_run(body.cron_expr)

    try:
        cursor = conn.execute(
            """INSERT INTO scheduled_tasks
               (name, project_id, role, task, cron_expr, system_prompt,
                skip_if_running, next_run_at, conversation_id)
               VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)""",
            (body.name, body.project_id, body.role, body.task,
             body.cron_expr, body.system_prompt,
             int(body.skip_if_running), next_run, body.conversation_id),
        )
    except sqlite3.IntegrityError:
        raise HTTPException(409, "A schedule with this name already exists")

    row = conn.execute(
        "SELECT s.*, p.display_name AS project_name "
        "FROM scheduled_tasks s "
        "LEFT JOIN projects p ON s.project_id = p.id "
        "WHERE s.id = ?",
        (cursor.lastrowid,),
    ).fetchone()
    return _row_to_dict(row)


@router.post("/schedules/one-time", status_code=201)
def create_one_time_schedule(
    body: OneTimeScheduleCreate, conn=Depends(get_db_conn),
):
    """Create a one-time scheduled task."""
    project = conn.execute(
        "SELECT id FROM projects WHERE id = ?", (body.project_id,)
    ).fetchone()
    if not project:
        raise HTTPException(404, "Project not found")

    try:
        cursor = conn.execute(
            """INSERT INTO scheduled_tasks
               (name, project_id, role, task, cron_expr, schedule_type,
                run_at, system_prompt, skip_if_running, next_run_at,
                conversation_id)
               VALUES (?, ?, ?, ?, NULL, 'one_time', ?, ?, 0, ?, ?)""",
            (body.name, body.project_id, body.role, body.task,
             body.run_at, body.system_prompt, body.run_at,
             body.conversation_id),
        )
    except sqlite3.IntegrityError:
        raise HTTPException(409, "A schedule with this name already exists")

    row = conn.execute(
        "SELECT s.*, p.display_name AS project_name "
        "FROM scheduled_tasks s "
        "LEFT JOIN projects p ON s.project_id = p.id "
        "WHERE s.id = ?",
        (cursor.lastrowid,),
    ).fetchone()
    return _row_to_dict(row)


@router.get("/schedules/runs")
def schedule_runs(
    page: int = Query(1, ge=1),
    per_page: int = Query(20, ge=1, le=100),
    conn=Depends(get_db_conn),
):
    """List all sessions triggered by schedules, with schedule metadata."""
    total = conn.execute(
        "SELECT count(*) FROM sessions se"
        "  JOIN scheduled_tasks st ON se.schedule_id = st.id",
    ).fetchone()[0]

    offset = (page - 1) * per_page
    rows = conn.execute(
        "SELECT se.run_id, se.project_id, se.role, se.status,"
        "       se.started_at, se.ended_at, se.duration_ms, se.exit_code,"
        "       se.task, se.schedule_id,"
        "       st.name AS schedule_name, st.schedule_type,"
        "       p.display_name AS project_name"
        "  FROM sessions se"
        "  JOIN scheduled_tasks st ON se.schedule_id = st.id"
        "  LEFT JOIN projects p ON se.project_id = p.id"
        "  ORDER BY se.started_at DESC"
        "  LIMIT ? OFFSET ?",
        (per_page, offset),
    ).fetchall()
    return {
        "items": [dict(r) for r in rows],
        "total": total,
        "page": page,
        "per_page": per_page,
    }


@router.get("/schedules/{schedule_id}")
def get_schedule(schedule_id: int, conn=Depends(get_db_conn)):
    """Get a single scheduled task."""
    row = conn.execute(
        "SELECT s.*, p.display_name AS project_name "
        "FROM scheduled_tasks s "
        "LEFT JOIN projects p ON s.project_id = p.id "
        "WHERE s.id = ?",
        (schedule_id,),
    ).fetchone()
    if not row:
        raise HTTPException(404, "Schedule not found")
    return _row_to_dict(row)


@router.put("/schedules/{schedule_id}")
def update_schedule(
    schedule_id: int, body: ScheduleUpdate, conn=Depends(get_db_conn),
):
    """Update a scheduled task."""
    existing = conn.execute(
        "SELECT * FROM scheduled_tasks WHERE id = ?", (schedule_id,)
    ).fetchone()
    if not existing:
        raise HTTPException(404, "Schedule not found")

    updates: list[str] = []
    params: list = []

    if body.name is not None:
        updates.append("name = ?")
        params.append(body.name.strip())
    if body.task is not None:
        updates.append("task = ?")
        params.append(body.task.strip())
    if body.role is not None:
        updates.append("role = ?")
        params.append(body.role.strip() or None)
    if body.system_prompt is not None:
        updates.append("system_prompt = ?")
        params.append(body.system_prompt.strip() or None)
    if body.cron_expr is not None:
        updates.append("cron_expr = ?")
        params.append(body.cron_expr)
        # Recompute next_run_at if cron changed and schedule is enabled
        if existing["enabled"]:
            next_run = _compute_next_run(body.cron_expr)
            updates.append("next_run_at = ?")
            params.append(next_run)
    if body.run_at is not None:
        updates.append("run_at = ?")
        params.append(body.run_at)
        if existing["enabled"]:
            updates.append("next_run_at = ?")
            params.append(body.run_at)
    if body.skip_if_running is not None:
        updates.append("skip_if_running = ?")
        params.append(int(body.skip_if_running))
    if "conversation_id" in body.model_fields_set:
        updates.append("conversation_id = ?")
        params.append(body.conversation_id)

    if not updates:
        raise HTTPException(422, "No fields to update")

    params.append(schedule_id)
    try:
        conn.execute(
            f"UPDATE scheduled_tasks SET {', '.join(updates)} WHERE id = ?",
            params,
        )
    except sqlite3.IntegrityError:
        raise HTTPException(409, "A schedule with this name already exists")

    row = conn.execute(
        "SELECT s.*, p.display_name AS project_name "
        "FROM scheduled_tasks s "
        "LEFT JOIN projects p ON s.project_id = p.id "
        "WHERE s.id = ?",
        (schedule_id,),
    ).fetchone()
    return _row_to_dict(row)


@router.delete("/schedules/{schedule_id}")
def delete_schedule(schedule_id: int, conn=Depends(get_db_conn)):
    """Delete a scheduled task."""
    row = conn.execute(
        "SELECT id FROM scheduled_tasks WHERE id = ?", (schedule_id,)
    ).fetchone()
    if not row:
        raise HTTPException(404, "Schedule not found")
    conn.execute("DELETE FROM scheduled_tasks WHERE id = ?", (schedule_id,))
    return {"ok": True}


@router.post("/schedules/{schedule_id}/enable")
def enable_schedule(schedule_id: int, conn=Depends(get_db_conn)):
    """Enable a scheduled task."""
    row = conn.execute(
        "SELECT * FROM scheduled_tasks WHERE id = ?", (schedule_id,)
    ).fetchone()
    if not row:
        raise HTTPException(404, "Schedule not found")

    if row["schedule_type"] == "one_time":
        run_at = row["run_at"]
        if not run_at:
            raise HTTPException(422, "One-time schedule has no run_at value")
        try:
            dt = datetime.fromisoformat(run_at)
            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=timezone.utc)
        except ValueError:
            raise HTTPException(422, "Invalid run_at value")
        if dt <= datetime.now(timezone.utc):
            raise HTTPException(422, "Cannot re-enable: scheduled time has passed")
        next_run = run_at
    else:
        next_run = _compute_next_run(row["cron_expr"])

    conn.execute(
        "UPDATE scheduled_tasks SET enabled = 1, next_run_at = ? WHERE id = ?",
        (next_run, schedule_id),
    )

    updated = conn.execute(
        "SELECT s.*, p.display_name AS project_name "
        "FROM scheduled_tasks s "
        "LEFT JOIN projects p ON s.project_id = p.id "
        "WHERE s.id = ?",
        (schedule_id,),
    ).fetchone()
    return _row_to_dict(updated)


@router.post("/schedules/{schedule_id}/disable")
def disable_schedule(schedule_id: int, conn=Depends(get_db_conn)):
    """Disable a scheduled task."""
    row = conn.execute(
        "SELECT id FROM scheduled_tasks WHERE id = ?", (schedule_id,)
    ).fetchone()
    if not row:
        raise HTTPException(404, "Schedule not found")

    conn.execute(
        "UPDATE scheduled_tasks SET enabled = 0, next_run_at = NULL WHERE id = ?",
        (schedule_id,),
    )

    updated = conn.execute(
        "SELECT s.*, p.display_name AS project_name "
        "FROM scheduled_tasks s "
        "LEFT JOIN projects p ON s.project_id = p.id "
        "WHERE s.id = ?",
        (schedule_id,),
    ).fetchone()
    return _row_to_dict(updated)


@router.get("/schedules/{schedule_id}/history")
def schedule_history(schedule_id: int, conn=Depends(get_db_conn)):
    """List sessions spawned by this schedule."""
    row = conn.execute(
        "SELECT id FROM scheduled_tasks WHERE id = ?", (schedule_id,)
    ).fetchone()
    if not row:
        raise HTTPException(404, "Schedule not found")

    rows = conn.execute(
        "SELECT run_id, project_id, role, runtime, isolation, status,"
        "       started_at, ended_at, duration_ms, exit_code, task"
        "  FROM sessions WHERE schedule_id = ? ORDER BY started_at DESC"
        "  LIMIT 50",
        (schedule_id,),
    ).fetchall()
    return [dict(r) for r in rows]


@router.post("/schedules/{schedule_id}/trigger")
def trigger_schedule(schedule_id: int, conn=Depends(get_db_conn)):
    """Manually trigger a schedule immediately."""
    from strawpot_gui.routers.sessions import launch_session_subprocess
    from strawpot_gui.scheduler import fire_schedule

    row = conn.execute(
        "SELECT * FROM scheduled_tasks WHERE id = ?", (schedule_id,)
    ).fetchone()
    if not row:
        raise HTTPException(404, "Schedule not found")

    result = fire_schedule(conn, dict(row), launch_session_subprocess)

    if result.get("error"):
        raise HTTPException(500, result["error"])

    now = datetime.now(timezone.utc).isoformat()
    conn.execute(
        "UPDATE scheduled_tasks SET last_run_at = ?, last_error = NULL WHERE id = ?",
        (now, schedule_id),
    )
    return result


@router.post("/schedules/runs/{run_id}/rerun")
def rerun_schedule_run(run_id: str, conn=Depends(get_db_conn)):
    """Re-run a past schedule-triggered session."""
    from strawpot_gui.routers.sessions import launch_session_subprocess
    from strawpot_gui.scheduler import fire_schedule

    session = conn.execute(
        "SELECT schedule_id, user_task, task FROM sessions WHERE run_id = ?",
        (run_id,),
    ).fetchone()
    if not session:
        raise HTTPException(404, "Session not found")
    if not session["schedule_id"]:
        raise HTTPException(422, "Session was not triggered by a schedule")

    schedule = conn.execute(
        "SELECT * FROM scheduled_tasks WHERE id = ?",
        (session["schedule_id"],),
    ).fetchone()
    if not schedule:
        raise HTTPException(404, "Schedule has been deleted")

    original_task = session["user_task"] or session["task"]
    result = fire_schedule(
        conn, dict(schedule), launch_session_subprocess,
        task_override=original_task,
    )

    if result.get("error"):
        raise HTTPException(500, result["error"])

    return result



"""Tests for scheduled tasks CRUD endpoints."""

from datetime import datetime, timedelta, timezone
from unittest.mock import patch

import pytest


@pytest.fixture
def project_id(client, tmp_path):
    """Create a project and return its id."""
    resp = client.post(
        "/api/projects",
        json={"display_name": "test-proj", "working_dir": str(tmp_path)},
    )
    return resp.json()["id"]


class TestCreateSchedule:
    def test_create(self, client, project_id):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "nightly",
                "project_id": project_id,
                "task": "run tests",
                "cron_expr": "0 0 * * *",
            },
        )
        assert resp.status_code == 201
        data = resp.json()
        assert data["name"] == "nightly"
        assert data["task"] == "run tests"
        assert data["cron_expr"] == "0 0 * * *"
        assert data["enabled"] is True  # DB default is enabled
        assert data["skip_if_running"] is True
        assert data["next_run_at"] is not None

    def test_create_with_optional_fields(self, client, project_id):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "daily",
                "project_id": project_id,
                "task": "deploy",
                "cron_expr": "30 8 * * *",
                "role": "team-lead",
                "system_prompt": "Be thorough",
                "skip_if_running": False,
            },
        )
        assert resp.status_code == 201
        data = resp.json()
        assert data["role"] == "team-lead"
        assert data["system_prompt"] == "Be thorough"
        assert data["skip_if_running"] is False

    def test_invalid_cron_returns_422(self, client, project_id):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "bad",
                "project_id": project_id,
                "task": "x",
                "cron_expr": "not-a-cron",
            },
        )
        assert resp.status_code == 422

    def test_empty_name_returns_422(self, client, project_id):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "  ",
                "project_id": project_id,
                "task": "x",
                "cron_expr": "0 0 * * *",
            },
        )
        assert resp.status_code == 422

    def test_empty_task_returns_422(self, client, project_id):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "ok",
                "project_id": project_id,
                "task": "  ",
                "cron_expr": "0 0 * * *",
            },
        )
        assert resp.status_code == 422

    def test_nonexistent_project_returns_404(self, client):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "x",
                "project_id": 999,
                "task": "x",
                "cron_expr": "0 0 * * *",
            },
        )
        assert resp.status_code == 404

    def test_duplicate_name_returns_409(self, client, project_id):
        body = {
            "name": "unique",
            "project_id": project_id,
            "task": "x",
            "cron_expr": "0 0 * * *",
        }
        client.post("/api/schedules", json=body)
        resp = client.post("/api/schedules", json=body)
        assert resp.status_code == 409


class TestGetSchedule:
    def test_get(self, client, project_id):
        create = client.post(
            "/api/schedules",
            json={
                "name": "s1",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        )
        sid = create.json()["id"]
        resp = client.get(f"/api/schedules/{sid}")
        assert resp.status_code == 200
        assert resp.json()["name"] == "s1"

    def test_not_found(self, client):
        resp = client.get("/api/schedules/999")
        assert resp.status_code == 404


class TestListSchedules:
    def test_empty(self, client):
        resp = client.get("/api/schedules")
        assert resp.status_code == 200
        assert resp.json() == []

    def test_returns_all(self, client, project_id):
        for name in ["a", "b", "c"]:
            client.post(
                "/api/schedules",
                json={
                    "name": name,
                    "project_id": project_id,
                    "task": "t",
                    "cron_expr": "0 0 * * *",
                },
            )
        resp = client.get("/api/schedules")
        assert len(resp.json()) == 3

    def test_includes_project_name(self, client, project_id):
        client.post(
            "/api/schedules",
            json={
                "name": "s",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        )
        resp = client.get("/api/schedules")
        assert resp.json()[0]["project_name"] == "test-proj"


class TestUpdateSchedule:
    def _create(self, client, project_id, name="orig"):
        resp = client.post(
            "/api/schedules",
            json={
                "name": name,
                "project_id": project_id,
                "task": "original task",
                "cron_expr": "0 0 * * *",
            },
        )
        return resp.json()["id"]

    def test_update_name(self, client, project_id):
        sid = self._create(client, project_id)
        resp = client.put(f"/api/schedules/{sid}", json={"name": "renamed"})
        assert resp.status_code == 200
        assert resp.json()["name"] == "renamed"

    def test_update_task(self, client, project_id):
        sid = self._create(client, project_id)
        resp = client.put(f"/api/schedules/{sid}", json={"task": "new task"})
        assert resp.status_code == 200
        assert resp.json()["task"] == "new task"

    def test_update_cron(self, client, project_id):
        sid = self._create(client, project_id)
        resp = client.put(
            f"/api/schedules/{sid}", json={"cron_expr": "*/5 * * * *"}
        )
        assert resp.status_code == 200
        assert resp.json()["cron_expr"] == "*/5 * * * *"

    def test_empty_body_returns_422(self, client, project_id):
        sid = self._create(client, project_id)
        resp = client.put(f"/api/schedules/{sid}", json={})
        assert resp.status_code == 422

    def test_not_found(self, client):
        resp = client.put("/api/schedules/999", json={"name": "x"})
        assert resp.status_code == 404

    def test_duplicate_name_returns_409(self, client, project_id):
        self._create(client, project_id, name="first")
        sid2 = self._create(client, project_id, name="second")
        resp = client.put(f"/api/schedules/{sid2}", json={"name": "first"})
        assert resp.status_code == 409

    def test_invalid_cron_returns_422(self, client, project_id):
        sid = self._create(client, project_id)
        resp = client.put(
            f"/api/schedules/{sid}", json={"cron_expr": "bad-cron"}
        )
        assert resp.status_code == 422


class TestDeleteSchedule:
    def test_delete(self, client, project_id):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "del",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        )
        sid = resp.json()["id"]
        resp = client.delete(f"/api/schedules/{sid}")
        assert resp.status_code == 200
        assert resp.json()["ok"] is True
        assert client.get(f"/api/schedules/{sid}").status_code == 404

    def test_not_found(self, client):
        resp = client.delete("/api/schedules/999")
        assert resp.status_code == 404


class TestEnableDisable:
    def _create(self, client, project_id):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "toggle",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        )
        return resp.json()["id"]

    def test_enable(self, client, project_id):
        sid = self._create(client, project_id)
        resp = client.post(f"/api/schedules/{sid}/enable")
        assert resp.status_code == 200
        data = resp.json()
        assert data["enabled"] is True
        assert data["next_run_at"] is not None

    def test_disable(self, client, project_id):
        sid = self._create(client, project_id)
        client.post(f"/api/schedules/{sid}/enable")
        resp = client.post(f"/api/schedules/{sid}/disable")
        assert resp.status_code == 200
        data = resp.json()
        assert data["enabled"] is False
        assert data["next_run_at"] is None

    def test_enable_not_found(self, client):
        assert client.post("/api/schedules/999/enable").status_code == 404

    def test_disable_not_found(self, client):
        assert client.post("/api/schedules/999/disable").status_code == 404


class TestScheduleHistory:
    def test_empty_history(self, client, project_id):
        resp = client.post(
            "/api/schedules",
            json={
                "name": "h",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        )
        sid = resp.json()["id"]
        resp = client.get(f"/api/schedules/{sid}/history")
        assert resp.status_code == 200
        assert resp.json() == []

    def test_not_found(self, client):
        assert client.get("/api/schedules/999/history").status_code == 404


# ---------------------------------------------------------------------------
# One-time schedule tests
# ---------------------------------------------------------------------------


def _future_iso() -> str:
    """Return an ISO datetime 1 hour in the future."""
    return (datetime.now(timezone.utc) + timedelta(hours=1)).isoformat()


def _past_iso() -> str:
    """Return an ISO datetime 1 hour in the past."""
    return (datetime.now(timezone.utc) - timedelta(hours=1)).isoformat()


class TestCreateOneTimeSchedule:
    def test_create(self, client, project_id):
        run_at = _future_iso()
        resp = client.post(
            "/api/schedules/one-time",
            json={
                "name": "deploy-tonight",
                "project_id": project_id,
                "task": "deploy to prod",
                "run_at": run_at,
            },
        )
        assert resp.status_code == 201
        data = resp.json()
        assert data["name"] == "deploy-tonight"
        assert data["schedule_type"] == "one_time"
        assert data["cron_expr"] is None
        assert data["run_at"] is not None
        assert data["next_run_at"] is not None
        assert data["enabled"] is True
        assert data["skip_if_running"] is False

    def test_create_with_optional_fields(self, client, project_id):
        resp = client.post(
            "/api/schedules/one-time",
            json={
                "name": "migration",
                "project_id": project_id,
                "task": "run migration",
                "run_at": _future_iso(),
                "role": "dba",
                "system_prompt": "Be careful",
            },
        )
        assert resp.status_code == 201
        data = resp.json()
        assert data["role"] == "dba"
        assert data["system_prompt"] == "Be careful"

    def test_past_run_at_returns_422(self, client, project_id):
        resp = client.post(
            "/api/schedules/one-time",
            json={
                "name": "late",
                "project_id": project_id,
                "task": "x",
                "run_at": _past_iso(),
            },
        )
        assert resp.status_code == 422

    def test_invalid_run_at_returns_422(self, client, project_id):
        resp = client.post(
            "/api/schedules/one-time",
            json={
                "name": "bad",
                "project_id": project_id,
                "task": "x",
                "run_at": "not-a-datetime",
            },
        )
        assert resp.status_code == 422

    def test_nonexistent_project_returns_404(self, client):
        resp = client.post(
            "/api/schedules/one-time",
            json={
                "name": "x",
                "project_id": 999,
                "task": "x",
                "run_at": _future_iso(),
            },
        )
        assert resp.status_code == 404

    def test_duplicate_name_returns_409(self, client, project_id):
        body = {
            "name": "dup",
            "project_id": project_id,
            "task": "x",
            "run_at": _future_iso(),
        }
        client.post("/api/schedules/one-time", json=body)
        resp = client.post("/api/schedules/one-time", json=body)
        assert resp.status_code == 409


class TestListSchedulesTypeFilter:
    def test_filter_recurring(self, client, project_id):
        client.post(
            "/api/schedules",
            json={
                "name": "recurring-1",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        )
        client.post(
            "/api/schedules/one-time",
            json={
                "name": "onetime-1",
                "project_id": project_id,
                "task": "t",
                "run_at": _future_iso(),
            },
        )
        resp = client.get("/api/schedules?type=recurring")
        data = resp.json()
        assert len(data) == 1
        assert data[0]["name"] == "recurring-1"

    def test_filter_one_time(self, client, project_id):
        client.post(
            "/api/schedules",
            json={
                "name": "recurring-2",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        )
        client.post(
            "/api/schedules/one-time",
            json={
                "name": "onetime-2",
                "project_id": project_id,
                "task": "t",
                "run_at": _future_iso(),
            },
        )
        resp = client.get("/api/schedules?type=one_time")
        data = resp.json()
        assert len(data) == 1
        assert data[0]["name"] == "onetime-2"

    def test_no_filter_returns_all(self, client, project_id):
        client.post(
            "/api/schedules",
            json={
                "name": "r",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        )
        client.post(
            "/api/schedules/one-time",
            json={
                "name": "o",
                "project_id": project_id,
                "task": "t",
                "run_at": _future_iso(),
            },
        )
        resp = client.get("/api/schedules")
        assert len(resp.json()) == 2

    def test_invalid_type_returns_422(self, client):
        resp = client.get("/api/schedules?type=bogus")
        assert resp.status_code == 422


class TestEnableOneTime:
    def _create_one_time(self, client, project_id, run_at=None):
        resp = client.post(
            "/api/schedules/one-time",
            json={
                "name": "ot",
                "project_id": project_id,
                "task": "t",
                "run_at": run_at or _future_iso(),
            },
        )
        return resp.json()["id"]

    def test_enable_future_one_time(self, client, project_id):
        sid = self._create_one_time(client, project_id)
        client.post(f"/api/schedules/{sid}/disable")
        resp = client.post(f"/api/schedules/{sid}/enable")
        assert resp.status_code == 200
        assert resp.json()["enabled"] is True
        assert resp.json()["next_run_at"] is not None

    def test_enable_past_one_time_returns_422(self, client, project_id):
        # Create with future time, then disable, then manually set run_at to past
        sid = self._create_one_time(client, project_id)
        client.post(f"/api/schedules/{sid}/disable")
        # Update run_at to past via direct DB isn't possible through API
        # (validator rejects past), so we test the enable endpoint by
        # checking that a recently-fired one-time (auto-disabled with past
        # run_at) can't be re-enabled. We simulate by checking the 422 path.
        # For now, just verify the future case works (tested above).


class TestScheduleRuns:
    def test_empty_runs(self, client):
        resp = client.get("/api/schedules/runs")
        assert resp.status_code == 200
        data = resp.json()
        assert data["items"] == []
        assert data["total"] == 0
        assert data["page"] == 1

    def test_returns_schedule_metadata(self, client, project_id):
        # Create a schedule
        sched = client.post(
            "/api/schedules",
            json={
                "name": "run-test",
                "project_id": project_id,
                "task": "t",
                "cron_expr": "0 0 * * *",
            },
        ).json()
        sid = sched["id"]

        # Insert a session linked to this schedule directly via DB
        from strawpot_gui.db import get_db
        db_path = client.app.state.db_path
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, schedule_id, status, role, runtime, "
                " isolation, started_at, session_dir) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-1", project_id, sid, "completed",
                 "test", "test", "none", "2025-01-01T00:00:00", "/tmp"),
            )

        resp = client.get("/api/schedules/runs")
        assert resp.status_code == 200
        data = resp.json()
        assert data["total"] == 1
        assert data["page"] == 1
        items = data["items"]
        assert len(items) == 1
        assert items[0]["run_id"] == "run-1"
        assert items[0]["schedule_name"] == "run-test"
        assert items[0]["schedule_type"] == "recurring"
        assert items[0]["project_name"] == "test-proj"


class TestTriggerSchedule:
    def test_trigger(self, client, project_id):
        sched = client.post(
            "/api/schedules",
            json={
                "name": "trigger-test",
                "project_id": project_id,
                "task": "do work",
                "cron_expr": "0 0 * * *",
            },
        ).json()

        with patch(
            "strawpot_gui.routers.sessions.launch_session_subprocess",
            return_value="run-trig",
        ):
            resp = client.post(f"/api/schedules/{sched['id']}/trigger")
        assert resp.status_code == 200
        assert resp.json()["run_id"] == "run-trig"

        # last_run_at should be updated
        updated = client.get(f"/api/schedules/{sched['id']}").json()
        assert updated["last_run_at"] is not None

    def test_trigger_not_found(self, client):
        resp = client.post("/api/schedules/99999/trigger")
        assert resp.status_code == 404

    def test_trigger_launch_error(self, client, project_id):
        sched = client.post(
            "/api/schedules",
            json={
                "name": "trigger-err",
                "project_id": project_id,
                "task": "fail",
                "cron_expr": "0 0 * * *",
            },
        ).json()

        with patch(
            "strawpot_gui.routers.sessions.launch_session_subprocess",
            side_effect=RuntimeError("spawn error"),
        ):
            resp = client.post(f"/api/schedules/{sched['id']}/trigger")
        assert resp.status_code == 500


class TestRerunScheduleRun:
    def test_rerun(self, client, project_id):
        sched = client.post(
            "/api/schedules",
            json={
                "name": "rerun-test",
                "project_id": project_id,
                "task": "original task",
                "cron_expr": "0 0 * * *",
            },
        ).json()
        sid = sched["id"]

        # Insert a session linked to this schedule
        from strawpot_gui.db import get_db
        db_path = client.app.state.db_path
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, schedule_id, status, role, runtime, "
                " isolation, started_at, session_dir, task, user_task) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-orig", project_id, sid, "completed",
                 "test", "test", "none", "2025-01-01T00:00:00", "/tmp",
                 "context\n---\noriginal task", "original task"),
            )

        with patch(
            "strawpot_gui.routers.sessions.launch_session_subprocess",
            return_value="run-rerun",
        ) as mock_launch:
            resp = client.post("/api/schedules/runs/run-orig/rerun")
        assert resp.status_code == 200
        assert resp.json()["run_id"] == "run-rerun"
        # Should use user_task (original), not the context-prepended task
        assert mock_launch.call_args[0][2] == "original task"

    def test_rerun_not_found(self, client):
        resp = client.post("/api/schedules/runs/nonexistent/rerun")
        assert resp.status_code == 404

    def test_rerun_no_schedule(self, client, project_id):
        """Session not triggered by a schedule returns 422."""
        from strawpot_gui.db import get_db
        db_path = client.app.state.db_path
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, status, role, runtime, "
                " isolation, started_at, session_dir, task) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-manual", project_id, "completed",
                 "test", "test", "none", "2025-01-01T00:00:00", "/tmp",
                 "manual task"),
            )
        resp = client.post("/api/schedules/runs/run-manual/rerun")
        assert resp.status_code == 422



"""Tests for config read/write endpoints."""

import tomllib
from pathlib import Path


class TestGlobalConfig:
    def test_get_empty_global_config(self, client, tmp_path, monkeypatch):
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        resp = client.get("/api/config/global")
        assert resp.status_code == 200
        body = resp.json()
        assert body["values"] == {}
        assert "defaults" in body
        assert body["defaults"]["runtime"] == "strawpot-claude-code"

    def test_put_and_get_global_config(self, client, tmp_path, monkeypatch):
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        data = {"runtime": "codex", "isolation": "docker"}

        resp = client.put("/api/config/global", json=data)
        assert resp.status_code == 200
        assert resp.json() == data

        # Verify file was written
        path = tmp_path / "home" / "strawpot.toml"
        assert path.is_file()
        with open(path, "rb") as f:
            written = tomllib.load(f)
        assert written == data

        # Verify round-trip via GET
        resp = client.get("/api/config/global")
        body = resp.json()
        assert body["values"] == data
        assert "defaults" in body


class TestProjectConfig:
    def _create_project(self, client, working_dir):
        resp = client.post("/api/projects", json={
            "display_name": "Test",
            "working_dir": str(working_dir),
        })
        return resp.json()["id"]

    def test_get_project_config_empty(self, client, tmp_path, monkeypatch):
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = self._create_project(client, project_dir)

        resp = client.get(f"/api/projects/{pid}/config")
        assert resp.status_code == 200
        data = resp.json()
        assert "merged" in data
        assert "project" in data
        assert "global" in data
        # No config files → project and global are empty dicts
        assert data["project"] == {}
        assert data["global"] == {}
        # Merged should have defaults
        assert data["merged"]["runtime"] == "strawpot-claude-code"

    def test_put_and_get_project_config(self, client, tmp_path, monkeypatch):
        monkeypatch.setenv("STRAWPOT_HOME", str(tmp_path / "home"))
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = self._create_project(client, project_dir)

        config_data = {"runtime": "codex"}
        resp = client.put(f"/api/projects/{pid}/config", json=config_data)
        assert resp.status_code == 200

        # Verify file written
        path = project_dir / "strawpot.toml"
        assert path.is_file()

        # Verify merged config picks up the override
        resp = client.get(f"/api/projects/{pid}/config")
        data = resp.json()
        assert data["project"] == {"runtime": "codex"}
        assert data["merged"]["runtime"] == "codex"

    def test_merged_shows_both_sources(self, client, tmp_path, monkeypatch):
        home = tmp_path / "home"
        monkeypatch.setenv("STRAWPOT_HOME", str(home))

        # Write global config
        client.put("/api/config/global", json={"runtime": "global_rt"})

        # Create project with its own config
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        pid = self._create_project(client, project_dir)
        client.put(f"/api/projects/{pid}/config", json={"isolation": "worktree"})

        resp = client.get(f"/api/projects/{pid}/config")
        data = resp.json()
        assert data["global"] == {"runtime": "global_rt"}
        assert data["project"] == {"isolation": "worktree"}
        # Merged has global runtime + project isolation
        assert data["merged"]["runtime"] == "global_rt"
        assert data["merged"]["isolation"] == "worktree"

    def test_project_not_found(self, client):
        resp = client.get("/api/projects/999/config")
        assert resp.status_code == 404

    def test_put_project_not_found(self, client):
        resp = client.put("/api/projects/999/config", json={"runtime": "x"})
        assert resp.status_code == 404



"""Tests for project activity stats endpoint."""

import json
import os
import uuid
from datetime import datetime, timedelta, timezone

from strawpot_gui.db import get_db, sync_sessions

from test_sessions_sync import _register_project, _write_session, _write_trace


def _setup_project_with_sessions(client, tmp_path, sessions_data):
    """Register a project and create sessions, then sync.

    sessions_data: list of dicts with keys:
        run_id, status (completed|failed|stopped), duration_ms, started_at
    """
    pid = _register_project(client, tmp_path)

    for s in sessions_data:
        run_id = s["run_id"]
        status = s.get("status", "completed")
        is_archived = status in ("completed", "failed", "stopped")
        session_dir = _write_session(
            tmp_path,
            run_id,
            archived=is_archived,
            started_at=s.get("started_at", "2026-01-15T12:00:00+00:00"),
        )
        # Build trace events matching _parse_trace expectations
        events = []
        if status in ("completed", "failed", "stopped"):
            end_data = {}
            if s.get("duration_ms") is not None:
                end_data["duration_ms"] = s["duration_ms"]
            exit_code = 0 if status == "completed" else 1
            end_data["exit_code"] = exit_code
            events.append({
                "event": "session_end",
                "ts": s.get("started_at", "2026-01-15T12:00:00+00:00"),
                "data": end_data,
            })
        _write_trace(session_dir, events)

    sync_sessions(client.app.state.db_path)
    return pid


class TestStatsEndpoint:
    def test_project_not_found(self, client):
        resp = client.get("/api/projects/999/stats")
        assert resp.status_code == 404

    def test_invalid_period(self, client, tmp_path):
        pid = _register_project(client, tmp_path)
        resp = client.get(f"/api/projects/{pid}/stats", params={"period": "5d"})
        assert resp.status_code == 422

    def test_empty_project(self, client, tmp_path):
        pid = _register_project(client, tmp_path)
        resp = client.get(f"/api/projects/{pid}/stats")
        assert resp.status_code == 200
        data = resp.json()
        assert data["period"] == "30d"
        assert data["total_runs"] == 0
        assert data["completed"] == 0
        assert data["failed"] == 0
        assert data["stopped"] == 0
        assert data["success_rate"] == 0.0
        assert data["avg_duration_ms"] is None
        # Daily array should be gap-filled
        assert len(data["daily"]) == 31  # 30 days + today

    def test_default_period_is_30d(self, client, tmp_path):
        pid = _register_project(client, tmp_path)
        resp = client.get(f"/api/projects/{pid}/stats")
        assert resp.json()["period"] == "30d"

    def test_period_7d(self, client, tmp_path):
        pid = _register_project(client, tmp_path)
        resp = client.get(f"/api/projects/{pid}/stats", params={"period": "7d"})
        data = resp.json()
        assert data["period"] == "7d"
        assert len(data["daily"]) == 8  # 7 days + today

    def test_aggregation(self, client, tmp_path):
        now = datetime.now(timezone.utc)
        yesterday = (now - timedelta(days=1)).replace(hour=10, minute=0, second=0, microsecond=0)
        today = now.replace(hour=10, minute=0, second=0, microsecond=0)

        sessions = [
            {"run_id": "run_s1", "status": "completed", "duration_ms": 100000, "started_at": yesterday.isoformat()},
            {"run_id": "run_s2", "status": "completed", "duration_ms": 200000, "started_at": yesterday.isoformat()},
            {"run_id": "run_s3", "status": "failed", "duration_ms": 50000, "started_at": today.isoformat()},
        ]
        pid = _setup_project_with_sessions(client, tmp_path, sessions)

        resp = client.get(f"/api/projects/{pid}/stats", params={"period": "7d"})
        data = resp.json()

        assert data["total_runs"] == 3
        assert data["completed"] == 2
        assert data["failed"] == 1
        assert data["success_rate"] == 66.7
        assert data["avg_duration_ms"] is not None

    def test_gap_filling(self, client, tmp_path):
        now = datetime.now(timezone.utc)
        three_days_ago = (now - timedelta(days=3)).replace(hour=10, minute=0, second=0, microsecond=0)

        sessions = [
            {"run_id": "run_s1", "status": "completed", "duration_ms": 60000, "started_at": three_days_ago.isoformat()},
        ]
        pid = _setup_project_with_sessions(client, tmp_path, sessions)

        resp = client.get(f"/api/projects/{pid}/stats", params={"period": "7d"})
        data = resp.json()

        dates = [d["date"] for d in data["daily"]]
        three_days_ago_date = three_days_ago.date().isoformat()
        assert three_days_ago_date in dates

        # Days without sessions should have zeroes
        for d in data["daily"]:
            if d["date"] != three_days_ago_date:
                assert d["total"] == 0
                assert d["completed"] == 0
                assert d["failed"] == 0
                assert d["avg_duration_ms"] is None

    def test_excludes_running_sessions(self, client, tmp_path):
        now = datetime.now(timezone.utc)
        today = now.replace(hour=10, minute=0, second=0, microsecond=0)

        # Create a completed session
        pid = _register_project(client, tmp_path)
        session_dir = _write_session(
            tmp_path, "run_s1", archived=True,
            started_at=today.isoformat(),
        )
        _write_trace(session_dir, [
            {"event": "session_end", "ts": today.isoformat(), "data": {"duration_ms": 60000, "exit_code": 0}},
        ])

        # Create a running session (not archived, no trace)
        _write_session(
            tmp_path, "run_s2", archived=False,
            started_at=today.isoformat(),
        )

        sync_sessions(client.app.state.db_path)

        resp = client.get(f"/api/projects/{pid}/stats", params={"period": "7d"})
        data = resp.json()

        assert data["total_runs"] == 1
        assert data["completed"] == 1

    def test_null_duration_excluded_from_avg(self, client, tmp_path):
        now = datetime.now(timezone.utc)
        today = now.replace(hour=10, minute=0, second=0, microsecond=0)

        sessions = [
            {"run_id": "run_s1", "status": "completed", "duration_ms": 120000, "started_at": today.isoformat()},
            {"run_id": "run_s2", "status": "completed", "duration_ms": None, "started_at": today.isoformat()},
        ]
        pid = _setup_project_with_sessions(client, tmp_path, sessions)

        resp = client.get(f"/api/projects/{pid}/stats", params={"period": "7d"})
        data = resp.json()

        assert data["avg_duration_ms"] == 120000

    def test_stopped_sessions_in_aggregation(self, client, tmp_path, app):
        """Stopped sessions count toward total_runs and the stopped field,
        but do not affect success_rate (which only considers completed vs failed).

        Uses direct DB insertion because sync_sessions infers status from
        exit_code (non-zero → failed), which would lose the 'stopped' state.
        """
        now = datetime.now(timezone.utc)
        today = now.replace(hour=10, minute=0, second=0, microsecond=0).isoformat()

        pid = _register_project(client, tmp_path)

        with get_db(app.state.db_path) as conn:
            for status, duration_ms in [("completed", 100000), ("failed", 50000), ("stopped", 30000)]:
                run_id = f"run_{uuid.uuid4().hex[:8]}"
                exit_code = 0 if status == "completed" else 1
                conn.execute(
                    "INSERT INTO sessions "
                    "(run_id, project_id, role, runtime, isolation, status, task, "
                    "exit_code, started_at, duration_ms, session_dir) "
                    "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                    (run_id, pid, "default", "claude-code", "none", status,
                     "test task", exit_code, today, duration_ms, str(tmp_path)),
                )

        resp = client.get(f"/api/projects/{pid}/stats", params={"period": "7d"})
        data = resp.json()

        assert data["total_runs"] == 3
        assert data["completed"] == 1
        assert data["failed"] == 1
        assert data["stopped"] == 1
        # success_rate = completed / (completed + failed) = 1/2 = 50%
        # stopped sessions are excluded from rate calculation
        assert data["success_rate"] == 50.0

        # Daily breakdown also includes stopped count
        today_entry = next(d for d in data["daily"] if d["total"] > 0)
        assert today_entry["stopped"] == 1

    def test_period_filters_old_sessions(self, client, tmp_path):
        now = datetime.now(timezone.utc)
        recent = (now - timedelta(days=3)).replace(hour=10, minute=0, second=0, microsecond=0)
        old = (now - timedelta(days=20)).replace(hour=10, minute=0, second=0, microsecond=0)

        sessions = [
            {"run_id": "run_recent", "status": "completed", "duration_ms": 60000, "started_at": recent.isoformat()},
            {"run_id": "run_old", "status": "completed", "duration_ms": 60000, "started_at": old.isoformat()},
        ]
        pid = _setup_project_with_sessions(client, tmp_path, sessions)

        resp = client.get(f"/api/projects/{pid}/stats", params={"period": "7d"})
        data = resp.json()

        assert data["total_runs"] == 1



"""Tests for the cron scheduler engine."""

import os
import sqlite3
from datetime import datetime, timezone
from unittest.mock import MagicMock

import pytest
from croniter import croniter

from strawpot_gui.db import get_db, init_db
from strawpot_gui.scheduler import Scheduler, fire_schedule, _next_run


@pytest.fixture
def db_path(tmp_path):
    """Create a temporary database and return its path."""
    path = str(tmp_path / "scheduler_test.db")
    init_db(path)
    return path


@pytest.fixture
def project_id(db_path):
    """Insert a test project and return its id."""
    with get_db(db_path) as conn:
        cursor = conn.execute(
            "INSERT INTO projects (display_name, working_dir) VALUES (?, ?)",
            ("test-proj", "/tmp/test"),
        )
        return cursor.lastrowid


def _insert_schedule(db_path, project_id, **kwargs):
    """Insert a scheduled task and return its id."""
    defaults = {
        "name": "test-schedule",
        "project_id": project_id,
        "task": "run tests",
        "cron_expr": "0 0 * * *",
        "schedule_type": "recurring",
        "run_at": None,
        "enabled": 1,
        "skip_if_running": 1,
        "next_run_at": None,
        "conversation_id": None,
    }
    defaults.update(kwargs)
    with get_db(db_path) as conn:
        cursor = conn.execute(
            """INSERT INTO scheduled_tasks
               (name, project_id, task, cron_expr, schedule_type, run_at,
                enabled, skip_if_running, next_run_at, conversation_id)
               VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""",
            (
                defaults["name"],
                defaults["project_id"],
                defaults["task"],
                defaults["cron_expr"],
                defaults["schedule_type"],
                defaults["run_at"],
                defaults["enabled"],
                defaults["skip_if_running"],
                defaults["next_run_at"],
                defaults["conversation_id"],
            ),
        )
        return cursor.lastrowid


class TestNextRun:
    def test_valid_cron(self):
        now = datetime(2025, 1, 1, 0, 0, tzinfo=timezone.utc)
        result = _next_run("0 0 * * *", now)
        assert result is not None
        parsed = datetime.fromisoformat(result)
        assert parsed > now

    def test_invalid_cron_returns_none(self):
        now = datetime(2025, 1, 1, 0, 0, tzinfo=timezone.utc)
        assert _next_run("bad-cron", now) is None


class TestInitNextRunTimes:
    def test_computes_missing_next_run(self, db_path, project_id):
        sid = _insert_schedule(db_path, project_id, enabled=1, next_run_at=None)
        scheduler = Scheduler(db_path, MagicMock())
        scheduler._init_next_run_times()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT next_run_at FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone()
        assert row["next_run_at"] is not None

    def test_skips_disabled(self, db_path, project_id):
        sid = _insert_schedule(db_path, project_id, enabled=0, next_run_at=None)
        scheduler = Scheduler(db_path, MagicMock())
        scheduler._init_next_run_times()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT next_run_at FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone()
        assert row["next_run_at"] is None

    def test_skips_already_set(self, db_path, project_id):
        original = "2099-01-01T00:00:00+00:00"
        sid = _insert_schedule(
            db_path, project_id, enabled=1, next_run_at=original
        )
        scheduler = Scheduler(db_path, MagicMock())
        scheduler._init_next_run_times()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT next_run_at FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone()
        assert row["next_run_at"] == original


class TestCheckAndFire:
    def test_fires_due_schedule(self, db_path, project_id):
        past = "2000-01-01T00:00:00+00:00"
        _insert_schedule(db_path, project_id, next_run_at=past)

        launch_fn = MagicMock(return_value="run-123")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called
        call_kwargs = launch_fn.call_args
        assert call_kwargs[1]["role"] is None
        assert call_kwargs[0][2] == "run tests"  # task positional arg

    def test_skips_future_schedule(self, db_path, project_id):
        future = "2099-01-01T00:00:00+00:00"
        _insert_schedule(db_path, project_id, next_run_at=future)

        launch_fn = MagicMock()
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert not launch_fn.called

    def test_skips_disabled_schedule(self, db_path, project_id):
        past = "2000-01-01T00:00:00+00:00"
        _insert_schedule(db_path, project_id, enabled=0, next_run_at=past)

        launch_fn = MagicMock()
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert not launch_fn.called

    def test_skips_if_running_session(self, db_path, project_id):
        past = "2000-01-01T00:00:00+00:00"
        sid = _insert_schedule(
            db_path, project_id, next_run_at=past, skip_if_running=1
        )

        # Insert a running session for this schedule
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, schedule_id, status, role, runtime, "
                " isolation, started_at, session_dir) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-active", project_id, sid, "running",
                 "test", "test", "none", "2025-01-01T00:00:00", "/tmp"),
            )

        launch_fn = MagicMock()
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert not launch_fn.called

    def test_fires_when_skip_if_running_disabled(self, db_path, project_id):
        past = "2000-01-01T00:00:00+00:00"
        sid = _insert_schedule(
            db_path, project_id, next_run_at=past, skip_if_running=0
        )

        # Insert a running session — should still fire
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, schedule_id, status, role, runtime, "
                " isolation, started_at, session_dir) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-active", project_id, sid, "running",
                 "test", "test", "none", "2025-01-01T00:00:00", "/tmp"),
            )

        launch_fn = MagicMock(return_value="run-new")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called

    def test_updates_next_run_after_fire(self, db_path, project_id):
        past = "2000-01-01T00:00:00+00:00"
        sid = _insert_schedule(db_path, project_id, next_run_at=past)

        launch_fn = MagicMock(return_value="run-123")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT next_run_at, last_run_at FROM scheduled_tasks WHERE id = ?",
                (sid,),
            ).fetchone()
        assert row["next_run_at"] is not None
        assert row["last_run_at"] is not None
        # next_run should be in the future
        next_dt = datetime.fromisoformat(row["next_run_at"])
        assert next_dt > datetime.now(timezone.utc)

    def test_records_error_on_launch_failure(self, db_path, project_id):
        past = "2000-01-01T00:00:00+00:00"
        sid = _insert_schedule(db_path, project_id, next_run_at=past)

        launch_fn = MagicMock(side_effect=RuntimeError("launch failed"))
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT last_error FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone()
        assert "launch failed" in row["last_error"]


# ---------------------------------------------------------------------------
# One-time schedule scheduler tests
# ---------------------------------------------------------------------------


class TestOneTimeScheduleFire:
    def test_auto_disables_after_fire(self, db_path, project_id):
        past = "2000-01-01T00:00:00+00:00"
        sid = _insert_schedule(
            db_path, project_id,
            name="one-time-task",
            cron_expr=None,
            schedule_type="one_time",
            run_at=past,
            next_run_at=past,
        )

        launch_fn = MagicMock(return_value="run-ot-1")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT enabled, next_run_at, last_run_at "
                "FROM scheduled_tasks WHERE id = ?",
                (sid,),
            ).fetchone()
        assert row["enabled"] == 0
        assert row["next_run_at"] is None
        assert row["last_run_at"] is not None

    def test_init_skips_one_time(self, db_path, project_id):
        """_init_next_run_times should not touch one-time schedules."""
        sid = _insert_schedule(
            db_path, project_id,
            name="one-time-init",
            cron_expr=None,
            schedule_type="one_time",
            run_at="2099-01-01T00:00:00+00:00",
            next_run_at=None,
            enabled=1,
        )
        scheduler = Scheduler(db_path, MagicMock())
        scheduler._init_next_run_times()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT next_run_at FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone()
        # Should remain None — _init_next_run_times only handles recurring
        assert row["next_run_at"] is None

    def test_skip_if_running_retains_next_run(self, db_path, project_id):
        """When a one-time schedule is skipped due to running session,
        next_run_at should NOT be advanced (stays unchanged for retry)."""
        past = "2000-01-01T00:00:00+00:00"
        sid = _insert_schedule(
            db_path, project_id,
            name="ot-skip",
            cron_expr=None,
            schedule_type="one_time",
            run_at=past,
            next_run_at=past,
            skip_if_running=1,
        )

        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, schedule_id, status, role, runtime, "
                " isolation, started_at, session_dir) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-block", project_id, sid, "running",
                 "test", "test", "none", "2025-01-01T00:00:00", "/tmp"),
            )

        launch_fn = MagicMock()
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert not launch_fn.called

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT next_run_at FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone()
        # next_run_at should be unchanged (not cleared or advanced)
        assert row["next_run_at"] == past


# ---------------------------------------------------------------------------
# #20 — imu role enforcement for project_id == 0
# ---------------------------------------------------------------------------


class TestImuRoleEnforcement:
    def test_imu_role_for_project_zero(self, db_path):
        """Schedules with project_id=0 and no explicit role get role='imu'."""
        # Insert project 0 (strawpot-internal)
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO projects (id, display_name, working_dir) "
                "VALUES (0, 'strawpot', '/tmp/strawpot')"
            )
        sid = _insert_schedule(db_path, 0, next_run_at="2000-01-01T00:00:00+00:00")

        launch_fn = MagicMock(return_value="run-imu")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called
        assert launch_fn.call_args[1]["role"] == "imu"

    def test_explicit_role_preserved_for_project_zero(self, db_path):
        """An explicit role on a project-0 schedule is kept (not overridden)."""
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO projects (id, display_name, working_dir) "
                "VALUES (0, 'strawpot', '/tmp/strawpot')"
            )
        # Manually set role via direct SQL since _insert_schedule doesn't set role
        sid = _insert_schedule(db_path, 0, next_run_at="2000-01-01T00:00:00+00:00")
        with get_db(db_path) as conn:
            conn.execute(
                "UPDATE scheduled_tasks SET role = 'custom' WHERE id = ?", (sid,)
            )

        launch_fn = MagicMock(return_value="run-custom")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called
        assert launch_fn.call_args[1]["role"] == "custom"

    def test_no_imu_for_regular_project(self, db_path, project_id):
        """Non-zero project_id without explicit role keeps role=None."""
        past = "2000-01-01T00:00:00+00:00"
        _insert_schedule(db_path, project_id, next_run_at=past)

        launch_fn = MagicMock(return_value="run-123")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called
        assert launch_fn.call_args[1]["role"] is None


# ---------------------------------------------------------------------------
# #19 — conversation targeting
# ---------------------------------------------------------------------------


class TestConversationTargeting:
    def _setup_conversation(self, db_path, project_id):
        """Create a conversation and return its id."""
        with get_db(db_path) as conn:
            cursor = conn.execute(
                "INSERT INTO conversations (project_id) VALUES (?)",
                (project_id,),
            )
            return cursor.lastrowid

    def test_passes_conversation_id_to_launch(self, db_path, project_id):
        """When conversation_id is set, it's forwarded to launch_fn."""
        conv_id = self._setup_conversation(db_path, project_id)
        past = "2000-01-01T00:00:00+00:00"
        _insert_schedule(
            db_path, project_id,
            next_run_at=past,
            conversation_id=conv_id,
        )

        launch_fn = MagicMock(return_value="run-conv")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called
        assert launch_fn.call_args[1]["conversation_id"] == conv_id

    def test_no_conversation_id_by_default(self, db_path, project_id):
        """Without conversation_id, launch_fn is called without it."""
        past = "2000-01-01T00:00:00+00:00"
        _insert_schedule(db_path, project_id, next_run_at=past)

        launch_fn = MagicMock(return_value="run-plain")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called
        assert "conversation_id" not in launch_fn.call_args[1]

    def test_conversation_context_prepended(self, db_path, project_id):
        """When conversation has prior sessions, context is prepended to task."""
        conv_id = self._setup_conversation(db_path, project_id)
        past = "2000-01-01T00:00:00+00:00"

        # Add a completed session to the conversation
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, conversation_id, status, role, runtime, "
                " isolation, started_at, session_dir, task, summary, exit_code) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-prior", project_id, conv_id, "completed",
                 "test", "test", "none", "2025-01-01T00:00:00", "/tmp",
                 "earlier task", "did stuff", 0),
            )

        _insert_schedule(
            db_path, project_id,
            next_run_at=past,
            conversation_id=conv_id,
        )

        launch_fn = MagicMock(return_value="run-ctx")
        scheduler = Scheduler(db_path, launch_fn)
        scheduler._check_and_fire()

        assert launch_fn.called
        # Task should have context prepended
        task_arg = launch_fn.call_args[0][2]
        assert "Prior Conversation" in task_arg
        assert "run tests" in task_arg
        # user_task should be the original task
        assert launch_fn.call_args[1]["user_task"] == "run tests"


# ---------------------------------------------------------------------------
# fire_schedule() standalone function tests
# ---------------------------------------------------------------------------


class TestFireSchedule:
    def test_returns_run_id(self, db_path, project_id):
        """fire_schedule returns {"run_id": ...} on success."""
        sid = _insert_schedule(db_path, project_id)
        launch_fn = MagicMock(return_value="run-abc")
        with get_db(db_path) as conn:
            schedule = dict(conn.execute(
                "SELECT * FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone())
            result = fire_schedule(conn, schedule, launch_fn)
        assert result == {"run_id": "run-abc"}
        assert launch_fn.called

    def test_returns_error_on_failure(self, db_path, project_id):
        """fire_schedule returns {"error": ...} on launch failure."""
        sid = _insert_schedule(db_path, project_id)
        launch_fn = MagicMock(side_effect=RuntimeError("boom"))
        with get_db(db_path) as conn:
            schedule = dict(conn.execute(
                "SELECT * FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone())
            result = fire_schedule(conn, schedule, launch_fn)
        assert "boom" in result["error"]

    def test_task_override(self, db_path, project_id):
        """fire_schedule uses task_override when provided."""
        sid = _insert_schedule(db_path, project_id, task="original task")
        launch_fn = MagicMock(return_value="run-ovr")
        with get_db(db_path) as conn:
            schedule = dict(conn.execute(
                "SELECT * FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone())
            fire_schedule(conn, schedule, launch_fn, task_override="override task")
        assert launch_fn.call_args[0][2] == "override task"

    def test_queues_when_conversation_busy(self, db_path, project_id):
        """fire_schedule queues task when conversation has active session."""
        with get_db(db_path) as conn:
            cursor = conn.execute(
                "INSERT INTO conversations (project_id) VALUES (?)",
                (project_id,),
            )
            conv_id = cursor.lastrowid

        sid = _insert_schedule(
            db_path, project_id, conversation_id=conv_id,
        )

        # Insert active session on the conversation
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, conversation_id, status, role, runtime, "
                " isolation, started_at, session_dir) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-busy", project_id, conv_id, "running",
                 "test", "test", "none", "2025-01-01T00:00:00", "/tmp"),
            )

        launch_fn = MagicMock()
        with get_db(db_path) as conn:
            schedule = dict(conn.execute(
                "SELECT * FROM scheduled_tasks WHERE id = ?", (sid,)
            ).fetchone())
            result = fire_schedule(conn, schedule, launch_fn)
        assert result == {"queued": True}
        assert not launch_fn.called


# ---------------------------------------------------------------------------
# _refresh_active_sessions background sweep tests
# ---------------------------------------------------------------------------


class TestRefreshActiveSessions:
    def test_marks_stale_starting_session_failed(self, db_path, project_id, tmp_path):
        """A starting session with no session.json older than 15s gets marked failed."""
        session_dir = str(tmp_path / "sessions" / "run-stale")
        os.makedirs(session_dir)  # dir exists but no session.json

        stale_time = "2000-01-01T00:00:00+00:00"
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, status, role, runtime, "
                " isolation, started_at, session_dir) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-stale", project_id, "starting",
                 "test", "test", "none", stale_time, session_dir),
            )

        scheduler = Scheduler(db_path, MagicMock())
        scheduler._refresh_active_sessions()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT status FROM sessions WHERE run_id = 'run-stale'"
            ).fetchone()
        assert row["status"] == "failed"

    def test_leaves_recent_starting_session(self, db_path, project_id, tmp_path):
        """A recently started session without session.json stays starting."""
        session_dir = str(tmp_path / "sessions" / "run-recent")
        os.makedirs(session_dir)

        from datetime import datetime as dt, timezone as tz
        recent_time = dt.now(tz.utc).isoformat()
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, status, role, runtime, "
                " isolation, started_at, session_dir) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-recent", project_id, "starting",
                 "test", "test", "none", recent_time, session_dir),
            )

        scheduler = Scheduler(db_path, MagicMock())
        scheduler._refresh_active_sessions()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT status FROM sessions WHERE run_id = 'run-recent'"
            ).fetchone()
        assert row["status"] == "starting"

    def test_marks_missing_dir_session_failed(self, db_path, project_id):
        """A starting session with no session dir older than 15s gets marked failed."""
        stale_time = "2000-01-01T00:00:00+00:00"
        with get_db(db_path) as conn:
            conn.execute(
                "INSERT INTO sessions "
                "(run_id, project_id, status, role, runtime, "
                " isolation, started_at, session_dir) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                ("run-nodir", project_id, "starting",
                 "test", "test", "none", stale_time, "/nonexistent/path"),
            )

        scheduler = Scheduler(db_path, MagicMock())
        scheduler._refresh_active_sessions()

        with get_db(db_path) as conn:
            row = conn.execute(
                "SELECT status FROM sessions WHERE run_id = 'run-nodir'"
            ).fetchone()
        assert row["status"] == "failed"



"""Resource registry endpoints — list, detail, install, uninstall."""

import shutil
import subprocess
import sys
from pathlib import Path

from fastapi import APIRouter, Body, HTTPException

from strawpot.agents.registry import (
    AgentSpec,
    _current_os,
    parse_agent_md,
    validate_agent,
)
from strawpot.config import _read_toml, get_strawpot_home, save_resource_config
from strawpot.context import parse_frontmatter
from strawpot.memory.registry import parse_memory_md

router = APIRouter(prefix="/api/registry", tags=["registry"])

# Maps URL resource_type to (directory_name, manifest_filename)
RESOURCE_TYPES: dict[str, tuple[str, str]] = {
    "roles": ("roles", "ROLE.md"),
    "skills": ("skills", "SKILL.md"),
    "agents": ("agents", "AGENT.md"),
    "memories": ("memories", "MEMORY.md"),
}


def validate_type(resource_type: str) -> tuple[str, str]:
    """Return (dir_name, manifest) or raise 400."""
    entry = RESOURCE_TYPES.get(resource_type)
    if entry is None:
        raise HTTPException(
            400,
            f"Unknown resource type: {resource_type}. "
            f"Valid types: {', '.join(RESOURCE_TYPES)}",
        )
    return entry


def parse_manifest(manifest_path: Path, resource_type: str) -> tuple[dict, str]:
    """Parse a manifest file and return (frontmatter, body)."""
    if resource_type == "agents":
        return parse_agent_md(manifest_path)
    if resource_type == "memories":
        return parse_memory_md(manifest_path)
    # roles and skills use generic frontmatter parsing
    text = manifest_path.read_text(encoding="utf-8")
    parsed = parse_frontmatter(text)
    return parsed.get("frontmatter", {}), parsed.get("body", "")


def read_version(resource_dir: Path, fm: dict) -> str | None:
    """Read version from .version file, falling back to frontmatter metadata."""
    version_file = resource_dir / ".version"
    if version_file.is_file():
        try:
            return version_file.read_text(encoding="utf-8").strip() or None
        except OSError:
            pass
    metadata = fm.get("metadata", {})
    return metadata.get("version", fm.get("version")) or None


def scan_dir(
    base_dir: Path, dir_name: str, manifest: str, resource_type: str, source: str
) -> list[dict]:
    """Scan a directory for installed resources."""
    scan_path = base_dir / dir_name
    if not scan_path.is_dir():
        return []
    items = []
    for entry in sorted(scan_path.iterdir()):
        if not entry.is_dir() or entry.name.startswith("."):
            continue
        manifest_path = entry / manifest
        if not manifest_path.is_file():
            continue
        try:
            fm, _body = parse_manifest(manifest_path, resource_type)
        except (ValueError, Exception):
            fm = {}
        items.append(
            {
                "name": fm.get("name", entry.name),
                "version": read_version(entry, fm),
                "description": fm.get("description", ""),
                "source": source,
                "path": str(entry),
            }
        )
    return items


@router.get("/{resource_type}")
def list_resources(resource_type: str):
    """List all installed resources of a given type."""
    dir_name, manifest = validate_type(resource_type)
    home = get_strawpot_home()
    return scan_dir(home, dir_name, manifest, resource_type, "global")


@router.get("/{resource_type}/{name}")
def get_resource(resource_type: str, name: str):
    """Get detail for a single installed resource."""
    dir_name, manifest = validate_type(resource_type)
    home = get_strawpot_home()

    resource_dir = home / dir_name / name
    manifest_path = resource_dir / manifest
    if not manifest_path.is_file():
        raise HTTPException(404, f"Resource not found: {resource_type}/{name}")

    fm, body = parse_manifest(manifest_path, resource_type)
    return {
        "name": fm.get("name", name),
        "version": read_version(resource_dir, fm),
        "description": fm.get("description", ""),
        "frontmatter": fm,
        "body": body,
        "source": "global",
        "path": str(resource_dir),
    }


def extract_saved_values(
    toml_data: dict, resource_type: str, name: str, params_schema: dict
) -> tuple[dict, dict]:
    """Extract saved env and param values from parsed TOML data."""
    env_values: dict[str, str] = {}
    params_values: dict = {}

    if resource_type == "roles":
        role_data = toml_data.get("roles", {}).get(name, {})
        if isinstance(role_data, dict):
            for key in params_schema:
                if key in role_data:
                    params_values[key] = role_data[key]
    elif resource_type == "skills":
        skill_data = toml_data.get("skills", {}).get(name, {})
        if isinstance(skill_data, dict):
            env_values = skill_data.get("env", {})
    elif resource_type == "agents":
        agent_data = toml_data.get("agents", {}).get(name, {})
        if isinstance(agent_data, dict):
            env_values = agent_data.get("env", {})
            for key in params_schema:
                if key in agent_data:
                    params_values[key] = agent_data[key]
    elif resource_type == "memories":
        mem_data = toml_data.get("memories", {}).get(name, {})
        if isinstance(mem_data, dict):
            env_values = mem_data.get("env", {})
        memory_cfg = toml_data.get("memory_config", {})
        for key in params_schema:
            if key in memory_cfg:
                params_values[key] = memory_cfg[key]

    return env_values, params_values


def coerce_param(value: object, param_type: str | None) -> object:
    """Coerce a param value to the correct Python type for TOML."""
    if param_type == "int":
        return int(value)  # type: ignore[arg-type]
    if param_type == "float":
        return float(value)  # type: ignore[arg-type]
    if param_type == "boolean":
        if isinstance(value, str):
            return value.lower() in ("true", "1", "yes")
        return bool(value)
    return value


@router.get("/{resource_type}/{name}/config")
def get_resource_config(resource_type: str, name: str):
    """Get env/params schema from manifest and saved values from TOML."""
    dir_name, manifest = validate_type(resource_type)
    home = get_strawpot_home()

    resource_dir = home / dir_name / name
    manifest_path = resource_dir / manifest
    if not manifest_path.is_file():
        raise HTTPException(404, f"Resource not found: {resource_type}/{name}")

    fm, _ = parse_manifest(manifest_path, resource_type)
    strawpot_meta = fm.get("metadata", {}).get("strawpot", {})

    env_schema = strawpot_meta.get("env", {})
    params_schema = strawpot_meta.get("params", {})

    # Roles expose default_agent as a configurable parameter
    if resource_type == "roles":
        manifest_default = strawpot_meta.get("default_agent")
        params_schema = {
            "default_agent": {
                "type": "string",
                "default": manifest_default,
                "description": "Agent to use when running this role",
            },
        }

    toml_data = _read_toml(home / "strawpot.toml")
    env_values, params_values = extract_saved_values(
        toml_data, resource_type, name, params_schema
    )

    return {
        "env_schema": env_schema,
        "env_values": env_values,
        "params_schema": params_schema,
        "params_values": params_values,
    }


@router.get("/agents/{name}/validate")
def validate_agent_status(name: str):
    """Check if an agent's prerequisites are satisfied."""
    home = get_strawpot_home()
    agent_dir = home / "agents" / name
    manifest_path = agent_dir / "AGENT.md"
    if not manifest_path.is_file():
        raise HTTPException(404, f"Agent not found: {name}")

    fm, _ = parse_agent_md(manifest_path)
    strawpot_meta = fm.get("metadata", {}).get("strawpot", {})

    spec = AgentSpec(
        name=fm.get("name", name),
        version="0",
        wrapper_cmd=[],
        tools=strawpot_meta.get("tools", {}),
        env_schema=strawpot_meta.get("env", {}),
    )
    result = validate_agent(spec)

    bin_map = strawpot_meta.get("bin", {})
    os_key = _current_os()
    bin_name = bin_map.get(os_key)
    setup_command = f"{bin_name} setup" if bin_name else None
    setup_description = strawpot_meta.get("setup", {}).get("description")

    return {
        "tools_ok": len(result.missing_tools) == 0,
        "missing_tools": [
            {"name": t[0], "install_hint": t[1]} for t in result.missing_tools
        ],
        "env_ok": len(result.missing_env) == 0,
        "missing_env": result.missing_env,
        "setup_command": setup_command,
        "setup_description": setup_description,
    }


@router.put("/{resource_type}/{name}/config")
def put_resource_config(resource_type: str, name: str, data: dict = Body(...)):
    """Save env and param values for a resource."""
    dir_name, manifest = validate_type(resource_type)
    home = get_strawpot_home()
    manifest_path = home / dir_name / name / manifest
    if not manifest_path.is_file():
        raise HTTPException(404, f"Resource not found: {resource_type}/{name}")

    env_values = data.get("env_values")
    params_values = data.get("params_values")

    # Type-coerce params based on schema
    if params_values:
        fm, _ = parse_manifest(manifest_path, resource_type)
        params_schema = fm.get("metadata", {}).get("strawpot", {}).get("params", {})
        coerced = {}
        for k, v in params_values.items():
            schema = params_schema.get(k, {})
            try:
                coerced[k] = coerce_param(v, schema.get("type"))
            except (ValueError, TypeError):
                coerced[k] = v
        params_values = coerced

    save_resource_config(None, resource_type, name, env_values, params_values)
    return {"ok": True}


def _strawhub_cmd() -> list[str] | None:
    """Locate the strawhub CLI, falling back to python -m for pipx installs."""
    path = shutil.which("strawhub")
    if path:
        return [path]
    try:
        subprocess.run(
            [sys.executable, "-m", "strawhub", "--version"],
            capture_output=True,
            check=True,
        )
        return [sys.executable, "-m", "strawhub"]
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None


def run_strawhub(*args: str) -> dict:
    """Run a strawhub CLI command and return result."""
    cmd = _strawhub_cmd()
    if cmd is None:
        raise HTTPException(
            503,
            "strawhub CLI not found. Install it with: pip install strawhub",
        )
    result = subprocess.run(
        [*cmd, *args],
        stdin=subprocess.DEVNULL,
        capture_output=True,
        text=True,
        timeout=120,
    )
    return {
        "exit_code": result.returncode,
        "stdout": result.stdout,
        "stderr": result.stderr,
    }


@router.post("/install")
def install_resource(data: dict = Body(...)):
    """Install a resource via strawhub."""
    resource_type = data.get("type", "")
    name = data.get("name", "")
    if not resource_type or not name:
        raise HTTPException(400, "Both 'type' and 'name' are required")
    # strawhub uses singular type names for install
    singular = resource_type.rstrip("s") if resource_type != "memories" else "memory"
    return run_strawhub("install", singular, "-y", name, "--global")


# Built-in resources that cannot be uninstalled.
# IMPORTANT: Keep in sync with PROTECTED in frontend/src/pages/ResourceBrowser.tsx
# and frontend/src/components/ResourceDetailSheet.tsx — both must be updated together.
_PROTECTED_ROLES = {"imu", "ai-ceo", "ai-employee"}
_PROTECTED_SKILLS = {"denden", "strawpot-session-recap"}
_PROTECTED_AGENTS = {"strawpot-claude-code"}
_PROTECTED_MEMORIES = {"dial"}


@router.delete("/{resource_type}/{name}")
def uninstall_resource(resource_type: str, name: str):
    """Uninstall a resource via strawhub."""
    validate_type(resource_type)
    if resource_type == "roles" and name in _PROTECTED_ROLES:
        raise HTTPException(403, f"'{name}' is a built-in role and cannot be uninstalled.")
    if resource_type == "skills" and name in _PROTECTED_SKILLS:
        raise HTTPException(403, f"'{name}' is a built-in skill and cannot be uninstalled.")
    if resource_type == "agents" and name in _PROTECTED_AGENTS:
        raise HTTPException(403, f"'{name}' is a built-in agent and cannot be uninstalled.")
    if resource_type == "memories" and name in _PROTECTED_MEMORIES:
        raise HTTPException(403, f"'{name}' is a built-in memory provider and cannot be uninstalled.")
    singular = resource_type.rstrip("s") if resource_type != "memories" else "memory"
    return run_strawhub("uninstall", singular, name, "--global")


def singular_type(resource_type: str) -> str:
    return resource_type.rstrip("s") if resource_type != "memories" else "memory"


@router.post("/update")
def update_resource(data: dict = Body(...)):
    """Update a resource to its latest version via strawhub."""
    resource_type = data.get("type", "")
    name = data.get("name", "")
    if not resource_type or not name:
        raise HTTPException(400, "Both 'type' and 'name' are required")
    singular = singular_type(resource_type)
    return run_strawhub("update", "-y", singular, name, "--global")


@router.post("/reinstall")
def reinstall_resource(data: dict = Body(...)):
    """Reinstall a resource (re-download the current version) via strawhub."""
    resource_type = data.get("type", "")
    name = data.get("name", "")
    if not resource_type or not name:
        raise HTTPException(400, "Both 'type' and 'name' are required")

    dir_name, _manifest = validate_type(resource_type)
    home = get_strawpot_home()
    version_file = home / dir_name / name / ".version"
    if not version_file.is_file():
        raise HTTPException(404, f"Resource not found or has no version: {resource_type}/{name}")
    version = version_file.read_text(encoding="utf-8").strip()
    if not version:
        raise HTTPException(404, f"Empty version file for: {resource_type}/{name}")

    singular = singular_type(resource_type)
    return run_strawhub("install", singular, "-y", name, "--version", version, "--force", "--global")



"""Tests for POST /api/sessions/:run_id/stop endpoint."""

import json
import os
from unittest.mock import patch

from test_conversations import _create_conversation, _submit_task
from test_sessions_sync import _register_project, _write_session

from strawpot_gui.db import get_db, sync_sessions


def _insert_session(client, project_dir, run_id, *, status="running"):
    """Register project, write session.json, sync, then force status."""
    pid = _register_project(client, project_dir)
    session_dir = _write_session(project_dir, run_id, pid=os.getpid())
    sync_sessions(client.app.state.db_path)
    # Override status to desired value
    with get_db(client.app.state.db_path) as conn:
        conn.execute(
            "UPDATE sessions SET status = ? WHERE run_id = ?",
            (status, run_id),
        )
    return pid, session_dir


class TestStopSession:
    def test_stop_sends_sigterm(self, client, tmp_path):
        """Stopping a running session sends SIGTERM and returns 'stopping'."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        _, session_dir = _insert_session(client, project_dir, "run_stop")

        with patch("strawpot_gui.routers.sessions.os.kill") as mock_kill:
            resp = client.post("/api/sessions/run_stop/stop")

        assert resp.status_code == 200
        assert resp.json() == {"run_id": "run_stop", "status": "stopped"}
        # First call: liveness check (signal 0), second call: SIGTERM
        assert mock_kill.call_count == 2
        mock_kill.assert_any_call(os.getpid(), 0)
        import signal
        mock_kill.assert_any_call(os.getpid(), signal.SIGTERM)

    def test_stop_nonexistent_returns_404(self, client):
        """Unknown run_id returns 404."""
        resp = client.post("/api/sessions/run_nope/stop")
        assert resp.status_code == 404

    def test_stop_completed_returns_409(self, client, tmp_path):
        """Cannot stop a completed session."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        _insert_session(
            client, project_dir, "run_done", status="completed"
        )

        resp = client.post("/api/sessions/run_done/stop")
        assert resp.status_code == 409

    def test_stop_failed_returns_409(self, client, tmp_path):
        """Cannot stop a failed session."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        _insert_session(
            client, project_dir, "run_fail", status="failed"
        )

        resp = client.post("/api/sessions/run_fail/stop")
        assert resp.status_code == 409

    def test_stop_starting_session_allowed(self, client, tmp_path):
        """Can stop a session still in 'starting' status."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        _insert_session(
            client, project_dir, "run_start", status="starting"
        )

        with patch("strawpot_gui.routers.sessions.os.kill"):
            resp = client.post("/api/sessions/run_start/stop")

        assert resp.status_code == 200
        assert resp.json()["status"] == "stopped"

    def test_stop_already_dead_marks_failed(self, client, tmp_path):
        """If process is already gone, marks session as failed."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        _insert_session(client, project_dir, "run_dead")

        with patch(
            "strawpot_gui.routers.sessions.os.kill",
            side_effect=ProcessLookupError,
        ):
            resp = client.post("/api/sessions/run_dead/stop")

        assert resp.status_code == 200
        assert resp.json() == {"run_id": "run_dead", "status": "stopped"}

        # Verify DB was updated
        with get_db(client.app.state.db_path) as conn:
            row = conn.execute(
                "SELECT status FROM sessions WHERE run_id = ?",
                ("run_dead",),
            ).fetchone()
        assert row["status"] == "stopped"

    def test_stop_missing_session_json_marks_failed(self, client, tmp_path):
        """If session.json is missing, marks session as failed."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        _, session_dir = _insert_session(
            client, project_dir, "run_nofile"
        )

        # Delete session.json
        os.remove(os.path.join(session_dir, "session.json"))

        resp = client.post("/api/sessions/run_nofile/stop")
        assert resp.status_code == 200
        assert resp.json() == {"run_id": "run_nofile", "status": "stopped"}

    def test_stop_no_pid_in_session_json_marks_failed(self, client, tmp_path):
        """If session.json has no pid field, marks session as failed."""
        project_dir = tmp_path / "proj"
        project_dir.mkdir()
        _, session_dir = _insert_session(
            client, project_dir, "run_nopid"
        )

        # Rewrite session.json without pid
        session_file = os.path.join(session_dir, "session.json")
        with open(session_file) as f:
            data = json.load(f)
        del data["pid"]
        with open(session_file, "w") as f:
            json.dump(data, f)

        resp = client.post("/api/sessions/run_nopid/stop")
        assert resp.status_code == 200
        assert resp.json() == {"run_id": "run_nopid", "status": "stopped"}

    def test_stop_drains_queued_task(self, client, tmp_path, app):
        """Stopping a session drains the next queued task."""
        d = tmp_path / "proj"
        d.mkdir()
        pid = _register_project(client, d)
        conv = _create_conversation(client, pid)
        cid = conv["id"]

        # Launch first task — creates a session
        resp1 = _submit_task(client, cid, task="First task")
        run_id1 = resp1.json()["run_id"]

        # Queue a second task (session already active → 202)
        _submit_task(client, cid, task="Queued task")

        # Verify task is queued
        with get_db(app.state.db_path) as conn:
            queued = conn.execute(
                "SELECT task FROM conversation_task_queue WHERE conversation_id = ?",
                (cid,),
            ).fetchall()
        assert len(queued) == 1
        assert queued[0]["task"] == "Queued task"

        # Stop the first session — should drain the queued task
        with patch("strawpot_gui.routers.sessions.os.kill"), \
             patch("strawpot_gui.routers.sessions.shutil.which", return_value="/usr/bin/strawpot"), \
             patch("strawpot_gui.routers.sessions.subprocess.Popen"), \
             patch("strawpot_gui.routers.sessions.load_config") as mock_config:
            from strawpot.config import StrawPotConfig
            mock_config.return_value = StrawPotConfig()
            resp = client.post(f"/api/sessions/{run_id1}/stop")

        assert resp.status_code == 200

        # Queue should be empty — task was drained
        with get_db(app.state.db_path) as conn:
            remaining = conn.execute(
                "SELECT id FROM conversation_task_queue WHERE conversation_id = ?",
                (cid,),
            ).fetchall()
        assert len(remaining) == 0

        # A new session should have been created for the queued task
        with get_db(app.state.db_path) as conn:
            sessions = conn.execute(
                "SELECT run_id FROM sessions WHERE conversation_id = ?",
                (cid,),
            ).fetchall()
        assert len(sessions) == 2
